from typing import Any, NamedTuple

from pydantic import (
    BaseModel as PydanticBaseModel,
)
from pydantic import (
    ConfigDict,
    Field,
    RootModel,
    model_validator,
)

from trac_rpc.validators import (
    OptionalTracDatetime,
    TracDatetime,
    TracOptionalField,
    TracSpaceSeparated,
    TracStrippedStr,
)

DEFAULT_CONFIG = ConfigDict(
    populate_by_name=True,
    validate_default=True,
    validate_assignment=True,
    frozen=True,
)


class BaseModel(PydanticBaseModel):
    model_config = DEFAULT_CONFIG


class TracRequest(BaseModel):
    id: int | None = None
    method: str
    params: list | None = None


class TracRpcErrorResponse(BaseModel):
    message: str
    code: int
    name: str


class TracResponse[ResponseT: Any](BaseModel):
    id: int | None
    error: TracRpcErrorResponse | None
    result: RootModel[ResponseT] | None

    @model_validator(mode="after")
    def check_non_overlapping_union_success_error(self) -> Any:
        if self.error is not None:
            if self.id is not None or self.result is not None:
                raise ValueError("invalid error response")
        elif self.result is None:
            raise ValueError("invalid success response")
        return self


class TracApiVersion(NamedTuple):
    epoch: int
    major: int
    minor: int


class TracComponent[StringType: str](BaseModel):
    name: StringType
    owner: StringType
    description: str


class TracMilestone[StringType: str](BaseModel):
    name: StringType
    description: str
    due: OptionalTracDatetime
    completed: OptionalTracDatetime


class TracVersion[StringType: str](BaseModel):
    name: StringType
    time: OptionalTracDatetime = Field(alias="released")
    description: str


class TracTicket[StringType: str](BaseModel):
    summary: StringType
    reporter: StringType
    owner: TracOptionalField[StringType]
    description: str
    type: StringType
    status: StringType
    priority: StringType
    milestone: TracOptionalField[StringType]
    component: StringType
    version: TracOptionalField[StringType]
    resolution: TracOptionalField[StringType]
    keywords: TracSpaceSeparated[str]
    cc: TracSpaceSeparated[str]
    time: TracDatetime
    changetime: TracDatetime


class TracTicketProperties[CustomTicketT: TracTicket](NamedTuple):
    id: int
    time_created: TracDatetime
    time_changed: TracDatetime
    attributes: CustomTicketT


class TracTicketChangelogEntry(NamedTuple):
    timestamp: TracDatetime
    author: TracOptionalField[TracStrippedStr]  # sometimes Trac doesn't record authorship!
    field: TracStrippedStr
    old_value: TracStrippedStr
    new_value: TracStrippedStr
    permanent: bool


TracTicketChangelog = list[TracTicketChangelogEntry]


class TracAttachment(NamedTuple):
    filename: TracStrippedStr
    description: TracStrippedStr
    size: int
    timestamp: TracDatetime
    author: TracStrippedStr


TracTicketAttachments = list[TracAttachment]



import json
from unittest.mock import MagicMock

import httpx
import pytest
import respx
from _pytest.monkeypatch import MonkeyPatch
from pydantic import ValidationError

from trac_rpc.client import ApiClient, HttpClient
from trac_rpc.exceptions import TracRpcError
from trac_rpc.models import TracApiVersion, TracRpcErrorResponse

from .utils import (
    RESPONSE_API_VERSION,
    TRAC_PASSWORD,
    TRAC_RPC_URL,
    TRAC_USERNAME,
    get_fixture,
)


def test_simple_init(api_client: ApiClient):
    assert api_client is not None


def test_auth_default(respx_mock: respx.mock):
    respx_mock.post(
        url=TRAC_RPC_URL,
        headers={
            "Authorization": "Basic dHJhY19sb2dpbjp0cmFjX3Bhc3N3b3Jk",
            "Content-Type": "application/json",
        },
    ).mock(return_value=RESPONSE_API_VERSION)

    api_client = ApiClient(
        rpc_url=TRAC_RPC_URL,
        http_client=HttpClient(auth=(TRAC_USERNAME, TRAC_PASSWORD)),
    )
    api_client.get_api_version()


def test_auth_override(respx_mock: respx.mock):
    respx_mock.post(
        url=TRAC_RPC_URL,
        headers={
            "X-Auth-Token": "1234",
            "Content-Type": "application/json",
        },
    ).mock(return_value=RESPONSE_API_VERSION)

    def custom_auth(request: httpx.Request) -> httpx.Request:
        request.headers["X-Auth-Token"] = "1234"
        return request

    api_client = ApiClient(rpc_url=TRAC_RPC_URL, http_client=HttpClient(auth=custom_auth))
    api_client.get_api_version()

    assert "Authorization" not in respx_mock.calls.last.request.headers


def test_event_hooks_default(monkeypatch: MonkeyPatch, respx_mock: respx.mock):
    respx_mock.post(
        url=TRAC_RPC_URL,
        headers={"Content-Type": "application/json"},
    ).mock(return_value=RESPONSE_API_VERSION)

    mock_log_request, mock_log_response = MagicMock(), MagicMock()
    monkeypatch.setattr(HttpClient, "log_trac_rpc_request", mock_log_request)
    monkeypatch.setattr(HttpClient, "log_trac_rpc_response", mock_log_response)

    api_client = ApiClient(rpc_url=TRAC_RPC_URL)
    api_client.get_api_version()

    mock_log_request.assert_called_once()
    mock_log_response.assert_called_once()


def test_raise_for_status(api_client: ApiClient, respx_mock: respx.mock):
    respx_mock.post(url=TRAC_RPC_URL) % httpx.codes.BAD_GATEWAY

    with pytest.raises(httpx.HTTPStatusError):
        api_client.get_api_version()


def test_event_hooks_override(monkeypatch: MonkeyPatch, respx_mock: respx.mock):
    respx_mock.post(TRAC_RPC_URL).mock(return_value=RESPONSE_API_VERSION)

    mock_log_request, mock_log_response = MagicMock(), MagicMock()
    monkeypatch.setattr(HttpClient, "log_trac_rpc_request", mock_log_request)
    monkeypatch.setattr(HttpClient, "log_trac_rpc_response", mock_log_response)

    api_client = ApiClient(rpc_url=TRAC_RPC_URL, http_client=HttpClient(event_hooks={}))
    api_client.get_api_version()

    mock_log_request.assert_not_called()
    mock_log_response.assert_not_called()


def test_raise_on_error(api_client: ApiClient, respx_mock: respx.mock):
    respx_mock.post(TRAC_RPC_URL).respond(
        status_code=httpx.codes.OK,  # sic!
        text=get_fixture("trac-response-rpc-error.json"),
    )

    with pytest.raises(TracRpcError) as excinfo:
        api_client.get_api_version()

    assert str(excinfo.value) == 'RPC method "ticket.wiki_to_html" not found'

    assert excinfo.value.error == TracRpcErrorResponse.model_validate(
        json.loads(get_fixture("trac-response-rpc-error.json"))["error"]
    )


def test_invalid_success_response(api_client: ApiClient, respx_mock: respx.mock):
    respx_mock.post().respond(text="""{"error": null, "result": null, "id": null}""")

    with pytest.raises(ValidationError, match=r".+ invalid success response .+"):
        api_client.get_api_version()


def test_invalid_error_response(api_client: ApiClient, respx_mock: respx.mock):
    respx_mock.post().respond(text="""{"error": {"message": "", "code": 123, "name": ""}, "result": null, "id": 123}""")

    with pytest.raises(ValidationError, match=r".+ invalid error response .+"):
        api_client.get_api_version()


def test_get_api_version(monkeypatch: MonkeyPatch, api_client, respx_mock: respx.mock):
    respx_mock.post(
        url=TRAC_RPC_URL,
        headers={"Content-Type": "application/json"},
    ).mock(return_value=RESPONSE_API_VERSION)

    response = api_client.get_api_version()

    assert isinstance(response, TracApiVersion)
    assert (response.epoch, response.major, response.minor) == (1, 1, 0)



import logging

import httpx

from trac_rpc.exceptions import TracRpcError
from trac_rpc.models import (
    TracApiVersion,
    TracComponent,
    TracMilestone,
    TracRequest,
    TracResponse,
    TracTicket,
    TracTicketAttachments,
    TracTicketChangelog,
    TracTicketChangelogEntry,
    TracTicketProperties,
    TracVersion,
)

logger = logging.getLogger(__name__)


class HttpClient(httpx.Client):
    @staticmethod
    def log_trac_rpc_request(request: httpx.Request):
        logger.debug(f"Trac API Request: {request.method} {request.url} {request.content.decode()}")

    @staticmethod
    def log_trac_rpc_response(response: httpx.Response):
        request = response.request
        response.read()
        logger.debug(f"Trac API Response: {request.method} {request.url} {response.status_code} {response.text}")

    def __init__(self, *, event_hooks=None, **kwargs):
        super().__init__(
            event_hooks=(
                event_hooks
                if event_hooks is not None
                else {
                    "request": [HttpClient.log_trac_rpc_request],
                    "response": [HttpClient.log_trac_rpc_response, httpx.Response.raise_for_status],
                }
            ),
            **kwargs,
        )


class ApiClient:
    def __init__(self, rpc_url: str, *, http_client: httpx.Client | None = None):
        self._rpc_url = rpc_url
        self._http_client = http_client if http_client is not None else HttpClient()

    def _request[T](self, request: TracRequest, klass: type[T]) -> T:
        http_response = self._http_client.post(self._rpc_url, json=request.model_dump())
        trac_response = TracResponse[klass].model_validate_json(http_response.text)

        if trac_response.error is not None:
            raise TracRpcError(trac_response.error.message, error=trac_response.error)

        return trac_response.result.root

    def _request_list_pod[T: str | int](self, function: str, klass: type[T]) -> list[T]:
        return self._request(TracRequest(method=function), list[klass])

    # system - Core of the RPC system
    def get_api_version(self) -> TracApiVersion:
        """
        Returns a list with three elements. First element is the epoch (0=Trac 0.10, 1=Trac 0.11 or higher). Second
        element is the major version number, third is the minor. Changes to the major version indicate API breaking
        changes, while minor version changes are simple additions, bug fixes, etc.
        """
        return self._request(TracRequest(method="system.getAPIVersion"), TracApiVersion)

    # ticket.component - Interface to ticket component objects
    # ticket.milestone - Interface to ticket milestone objects
    # ticket.priority - Interface to ticket priority
    # ticket.resolution - Interface to ticket resolution
    # ticket.severity - Interface to ticket severity
    # ticket.status - An interface to Trac ticket status objects (deprecated)
    # ticket.type - Interface to ticket type
    # ticket.version - Interface to ticket version objects

    def get_all_components(self) -> list[str]:
        """Get a list of all ticket component names"""
        return self._request_list_pod("ticket.component.getAll", str)

    def get_all_milestones(self) -> list[str]:
        """Get a list of all ticket milestone names"""
        return self._request_list_pod("ticket.milestone.getAll", str)

    def get_all_priorities(self) -> list[str]:
        """Get a list of all ticket priority names"""
        return self._request_list_pod("ticket.priority.getAll", str)

    def get_all_resolutions(self) -> list[str]:
        """Get a list of all ticket resolution names"""
        return self._request_list_pod("ticket.resolution.getAll", str)

    def get_all_severities(self) -> list[str]:
        """Get a list of all ticket severity names"""
        return self._request_list_pod("ticket.severity.getAll", str)

    def get_all_statuses(self) -> list[str]:
        """Returns all ticket states described by active workflow"""
        return self._request_list_pod("ticket.status.getAll", str)

    def get_all_types(self) -> list[str]:
        """Get a list of all ticket type names"""
        return self._request_list_pod("ticket.type.getAll", str)

    def get_all_versions(self) -> list[str]:
        """Get a list of all ticket version names"""
        return self._request_list_pod("ticket.version.getAll", str)

    def get_component[T: TracComponent](self, component_name: str, klass: type[T] = TracComponent) -> T:
        """Get a ticket component"""
        return self._request(TracRequest(method="ticket.component.get", params=[component_name]), klass)

    def get_milestone[T: TracMilestone](self, milestone_name: str, klass: type[T] = TracMilestone) -> T:
        """Get a ticket milestone"""
        return self._request(TracRequest(method="ticket.milestone.get", params=[milestone_name]), klass)

    def get_version[T: TracVersion](self, version_name: str, klass: type[T] = TracVersion) -> T:
        """Get a ticket priority"""
        return self._request(TracRequest(method="ticket.version.get", params=[version_name]), klass)

    # The following objects are just strings and do not need custom getters:
    #
    #   * priority
    #   * resolution
    #   * severity
    #   * status (deprecated)
    #   * type
    #

    # ticket - An interface to Trac's ticketing system
    def query_tickets(self, query: str = "", per_page: int = 0, page_number: int | None = None) -> list[int]:
        """
        Perform a ticket query, returning a list of ticket ID's. All queries will use stored settings for maximum
        number of results per page and paging options. Use max=n to define number of results to receive,
        and use page=n to page through larger result sets. Using max=0 will turn off paging and return all results.
        """
        pieces = (
            *((query,) if query != "" else ()),
            *((f"page={page_number}",) if page_number is not None else ()),
            f"max={per_page}",
        )
        return self._request(TracRequest(method="ticket.query", params=["&".join(pieces)]), list[int])

    def get_ticket_attachments(self, ticket_id: int) -> TracTicketAttachments:
        """
        Lists attachments for a given ticket. Returns (filename, description, size, time, author) for each attachment
        """
        return self._request(TracRequest(method="ticket.listAttachments", params=[ticket_id]), TracTicketAttachments)

    def get_ticket_changelog(self, ticket_id: int) -> TracTicketChangelog:
        """
        Return the changelog as a list of tuples of the form (time, author, field, oldvalue, newvalue, permanent).
        While the other tuple elements are quite self-explanatory, the permanent flag is used to distinguish
        collateral changes that are not yet immutable (like attachments, currently)
        """
        return self._request(TracRequest(method="ticket.changeLog", params=[ticket_id]), TracTicketChangelog)

    def get_ticket[T: TracTicket](self, ticket_id: int, klass: type[T] = TracTicket) -> TracTicketProperties[T]:
        """Fetch a ticket. Returns [id, time_created, time_changed, attributes]"""
        return self._request(TracRequest(method="ticket.get", params=[ticket_id]), TracTicketProperties[klass])

    # wiki - Superset of the WikiRPC API
    def get_all_wiki_pages(self) -> list[str]:
        """Returns a list of all pages. The result is an array of utf8 page names"""
        return self._request_list_pod("wiki.getAllPages", str)

    def wiki_to_html(self, text: str) -> str:
        """
        Render arbitrary Wiki text as HTML.

        For some migrations, the RPC server may need to be patched to force `escape_newlines=True`.
        """
        return self._request(TracRequest(method="wiki.wikiToHtml", params=[text]), str)

    # Utilities
    def get_ticket_last_field_change(
        self, ticket_id: int, field_name: str, new_value: str | None = None
    ) -> TracTicketChangelogEntry | None:
        """Get the changelog entry that represents the last time a given field was set to a particular value"""
        entry, *_ = [
            change
            for change in reversed(self.get_ticket_changelog(ticket_id))
            if change.field == field_name and (change.new_value == new_value if new_value is not None else True)
        ] + [None]
        return entry



import re
from typing import Annotated, Any

from pydantic import AfterValidator, AwareDatetime, BeforeValidator


def validate_datetime(value: Any) -> Any:
    if isinstance(value, dict):
        value_type, value_object = value["__jsonclass__"]
        if value_type == "datetime":
            return f"{value_object}Z"
    return value


def validate_optional_datetime(value: Any) -> Any:
    if value == 0:
        return None
    return validate_datetime(value)


def split_by_separator(separator: str, value: Any) -> Any:
    if isinstance(value, str):
        return [value.strip() for value in value.split(separator)] if value.strip() else []
    return value


def validate_space_separated(value: Any) -> Any:
    if isinstance(value, str):
        return split_by_separator(" ", re.sub(r" +", " ", value.strip()))
    return value


def validate_comma_separated(value: Any) -> Any:
    return split_by_separator(",", value)


def validate_space_or_comma_separated(value: Any) -> Any:
    if isinstance(value, str):
        return validate_comma_separated(value) if "," in value else validate_space_separated(value)
    return value


def validator_string_empty_to_none(v: Any) -> Any | None:
    if isinstance(v, str) and v.strip() == "":
        return None
    return v


def validator_string_strip(v: str) -> str:
    return v.strip()


def validate_in_set[T](value: T | None, allowed: set[T], optional: bool = False) -> T | None:
    if optional and value is None:
        return value
    if value not in allowed:
        raise ValueError(f"'{value}' is not in {allowed}")
    return value


type TracOptionalField[T: str | int] = Annotated[T | None, BeforeValidator(validator_string_empty_to_none)]

type TracDatetime = Annotated[AwareDatetime, BeforeValidator(validate_datetime)]
type OptionalTracDatetime = Annotated[AwareDatetime | None, BeforeValidator(validate_optional_datetime)]

type TracCommaSeparated[T: str] = Annotated[list[T], BeforeValidator(validate_comma_separated)]
type TracSpaceSeparated[T: str] = Annotated[list[T], BeforeValidator(validate_space_separated)]
type TracSpaceOrCommaSeparated[T: str] = Annotated[list[T], BeforeValidator(validate_space_or_comma_separated)]

type TracStrippedStr = Annotated[str, AfterValidator(validator_string_strip)]



import respx

from trac_rpc.client import ApiClient
from trac_rpc.models import TracMilestone
from trac_rpc.validators import TracStrippedStr

from .utils import get_fixture


def test_custom_string_type(api_client: ApiClient, respx_mock: respx.mock):
    respx_mock.post().respond(text=get_fixture("trac-get-milestone-response.json"))

    milestone = api_client.get_milestone(" milestone2  ", TracMilestone[TracStrippedStr])
    assert milestone.name == "milestone2"
